# -*- coding: utf-8 -*-

import matplotlib.pyplot as plt
import seaborn as sn
from pandas import DataFrame
from sklearn.feature_extraction.text import CountVectorizer
from sklearn.metrics import confusion_matrix, accuracy_score
from sklearn.model_selection import cross_validate, GridSearchCV
from sklearn.naive_bayes import MultinomialNB
from sklearn.utils.multiclass import unique_labels

# Pour le rapport

training_questions_fn = "./data/questions-t3.txt"
test_questions_fn = "./data/test-questions-t3.txt"

# Cette fonction provient de https://stackoverflow.com/a/7716358
def mean(numbers):
    return float(sum(numbers)) / max(len(numbers), 1)


def run_question_classification(training_fn, test_fn):
    accuracy_train, accuracy_test = train_and_test_classifier(training_fn, test_fn)
    print("Accuracy on training set: {0:.4f}".format(accuracy_train))
    print("Accuracy on test set: {0:.4f}".format(accuracy_test))


def custom_count_vectorize_questions(min_df, max_df, use_stop_words):
    stop_words = [None, 'english'][use_stop_words]
    count_vec = CountVectorizer(analyzer='word',
                                min_df=min_df,
                                max_df=max_df,
                                stop_words=stop_words)
    return count_vec


def train_and_test_classifier(training_fn, test_fn):
    questions, labels = load_dataset(training_fn)
    # questions, labels = load_dataset(training_questions_fn)
    print("Nb questions d'entraînement:", len(questions))
    test_questions, test_labels = load_dataset(test_fn)
    # test_questions, test_labels = load_dataset(test_questions_fn)
    print("Nb questions de test:", len(test_questions))

    # Insérer ici votre code pour la classification des questions.
    # Votre code...

    mnb = MultinomialNB()

    # Sélection des meilleurs paramètres pour le CountVectorizer
    # par recherche sur une grille et validation croisée
    # sur le modèle multinomial Naive Bayes

    avg_score = {}
    for min_df in range(1, 21):
        for max_df in [x / 200.0 for x in range(1, 21)]:
            for use_stop_words in [True, False]:
                count_vec = custom_count_vectorize_questions(min_df, max_df, use_stop_words)
                count_questions = count_vec.fit_transform(questions)
                mnb_fit = cross_validate(mnb, count_questions, labels, cv=5)
                avg_score[(min_df, max_df, use_stop_words)] = mean(mnb_fit['test_score'])

    # Meilleurs hyper-paramètres pour le CountVectorizer
    min_df_opt, max_df_opt, use_stop_words_opt = max(avg_score, key=avg_score.get)

    print("\nLes paramètres optimaux pour le CountVectorizer sont:\n" +
          'min_df: ' + str(min_df_opt) + "\n" +
          'max_df: ' + str(max_df_opt) + "\n" +
          'use_stop_words: ' + str(use_stop_words_opt))

    # On peut ainsi utiliser le meilleur CountVectorizer pour préparer les données pour le modèle Naive Bayes

    count_vec_opt = custom_count_vectorize_questions(min_df_opt, max_df_opt, use_stop_words_opt)
    count_questions_train_opt = count_vec_opt.fit_transform(questions)
    count_questions_test_opt = count_vec_opt.transform(test_questions)

    print("\nRecherche de l'hyperparamètre alpha:")

    gvc = GridSearchCV(mnb, param_grid={'alpha': [x / 10.0 for x in range(1, 21)]}, cv=5)

    mnb_fit = gvc.fit(count_questions_train_opt, labels)

    print("Score: {0:.4f}".format(str(mnb_fit.best_score_)))

    print("Paramètres: " + str(mnb_fit.get_params()))

    u_labels = unique_labels(labels, test_labels)

    labels_predict_train = mnb_fit.predict(count_questions_train_opt)
    labels_predict_test = mnb_fit.predict(count_questions_test_opt)

    accuracy_train = accuracy_score(y_true=labels, y_pred=labels_predict_train)
    accuracy_test = accuracy_score(y_true=test_labels, y_pred=labels_predict_test)

    # Section pour produire des images pour le rapport

    # Décompte des étiquettes du jeu d'entrainement

    df_count_labels = DataFrame(labels, columns=['category'])
    f, ax3 = plt.subplots(figsize=(9, 9))
    sn.set(style="darkgrid")
    sn_plot_count_category = sn.countplot(y='category', data=df_count_labels, palette="deep", ax=ax3, order=u_labels)
    sn_plot_count_category.set(title="Fréquence des catégories de questions\ndans l'échantillon d'entraînement",
                               xlabel="Fréquence",
                               ylabel="Catégorie")
    plt.tight_layout()
    sn_plot_count_category.get_figure().savefig("images/3_sn_plot_count_category_train.png")

    # Décompte des étiquettes du jeu de test

    df_count_test_labels = DataFrame(test_labels, columns=['category'])
    f, ax4 = plt.subplots(figsize=(9, 9))
    sn.reset_defaults()
    sn.set(style="darkgrid")
    sn_plot_count_category_test = sn.countplot(y='category', data=df_count_test_labels, palette="deep", ax=ax4, order=u_labels)
    sn_plot_count_category_test.set(title="Fréquence des catégories de questions\ndans l'échantillon de test",
                                    xlabel="Fréquence",
                                    ylabel="Catégorie")
    plt.tight_layout()
    sn_plot_count_category_test.get_figure().savefig("images/3_sn_plot_count_category_test.png")

    # Matrice de confusion pour l'entrainement
    cm_predict_train = confusion_matrix(y_true=labels, y_pred=labels_predict_train, labels=u_labels)
    df_cm_predict_train = DataFrame(cm_predict_train, index=u_labels, columns=u_labels)
    f, ax = plt.subplots(figsize=(9, 9))
    plt.subplots_adjust(top=0.961,
                        bottom=0.169,
                        left=0.169,
                        right=0.983,
                        hspace=0.2,
                        wspace=0.2)
    sn_plot_cm_predict_train = sn.heatmap(df_cm_predict_train,
                                          cmap='Oranges',
                                          annot=True,
                                          fmt="d",
                                          ax=ax)
    sn_plot_cm_predict_train.set(title="Performance du modèle NB sur l'échantillon d'entrainement",
                                 xlabel="Prédit",
                                 ylabel="Observé")
    sn_plot_cm_predict_train.get_figure().savefig("images/3_seaborn_df_cm_predict_train.png")

    # Matrice de confusion pour le test
    cm_predict_test = confusion_matrix(y_true=test_labels, y_pred=labels_predict_test, labels=u_labels)
    df_cm_predict_test = DataFrame(cm_predict_test, index=u_labels, columns=u_labels)
    f, ax2 = plt.subplots(figsize=(9, 9))
    plt.subplots_adjust(top=0.961,
                        bottom=0.169,
                        left=0.169,
                        right=0.983,
                        hspace=0.2,
                        wspace=0.2)
    sn_plot_cm_predict_test = sn.heatmap(df_cm_predict_test,
                                         cmap='Oranges',
                                         annot=True,
                                         fmt="d",
                                         ax=ax2)
    sn_plot_cm_predict_test.set(title="Performance du modèle NB sur l'échantillon test",
                                xlabel="Prédit",
                                ylabel="Observé")
    sn_plot_cm_predict_test.get_figure().savefig("images/3_seaborn_df_cm_predict_test.png")

    return accuracy_train, accuracy_test


def load_dataset(filename):
    with open(filename) as f:
        lines = f.read().splitlines()
        labels, questions = zip(*[tuple(s.split(' ', 1)) for s in lines])
    return questions, labels


if __name__ == '__main__':
    run_question_classification(training_questions_fn, test_questions_fn)
